/**
 * This file is part of pppBenchmark, a program accompanying
 * the paper: Abeel, T, Van de Peer Y, Saeys Y.  2009.  Towards a gold standard for promoter prediction evaluation. Bioinformatics. 25(12):i313-i320.
    * 
 * Copyright 2009 Thomas Abeel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 *
 */
package be.abeel.pppbenchmark;

import jargs.gnu.CmdLineParser.IllegalOptionValueException;
import jargs.gnu.CmdLineParser.Option;
import jargs.gnu.CmdLineParser.UnknownOptionException;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.SortedMap;
import java.util.TreeMap;

import org.jfree.chart.ChartFactory;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.xy.XYLineAndShapeRenderer;
import org.jfree.chart.title.LegendTitle;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;

import be.abeel.graphics.GraphicsFileExport;
import be.abeel.io.LineIterator;
import be.abeel.io.LineIteratorFactory;
import be.abeel.jargs.AutoHelpCmdLineParser;
import be.abeel.jfreechart.JFreeChartWrapper;

public class pppBenchmark {

	public static void main(String[] args) throws Exception {
		AutoHelpCmdLineParser parser = new AutoHelpCmdLineParser();

		Option geneO = parser.addHelp(parser.addStringOption("gene"), "<file> Use this set of genes for the B protocols. The specified file should contains the genes in GFF format. ");
		Option tagsO = parser.addHelp(parser.addStringOption("tags"), "<file> Use this set of tags for the A protocols. The specified file should contains the genes in GFF format. ");
		Option prefixO = parser.addHelp(parser.addStringOption("prefix"), "Add a prefix to all files that are outputted.");

		parse(parser, args);

		if (parser.checkHelp()) {
			System.exit(0);
		}
		String tmp = (String) parser.getOptionValue(geneO);
		if (tmp != null)
			Config.setGene(new File(tmp));

		tmp = (String) parser.getOptionValue(tagsO);
		if (tmp != null)
			Config.setTags(new File(tmp));

		// Boolean raw = (Boolean) parser.getOptionValue(rawO, Boolean.FALSE);
		// if (raw)
		// System.out.println("Outputting raw results");

		String prefix = (String) parser.getOptionValue(prefixO, "prefix");

		if (args.length == 0) {
			System.out.println("Usage: java -jar pppbenchmark-<version>.jar <file> <file> ...");
			System.out.println("or: java -jar pppbenchmark-<version>.jar --help for more details");
			System.out.println("You can specify as may files as you like. pppBenchmark will generate one report which will include the results from all prediction files.");
			System.out.println("The prediction files provided should be in a four column format: <chr> <start> <end> <score>.");
			System.out.println("The coordinates in the file should be relative to human genome release hg18.");
			System.out.println("If you need a different organism loaded, or if you have questions, send an e-mail to thomas@abeel.be");
			System.out.println();
			System.out.println("Copyright 2009 - Thomas Abeel");
			System.out.println();
			System.out
					.println("pppBenchmark is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.");
			System.out.println();
			System.out
					.println("pppBenchmark is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.");
			System.out.println();
			System.out.println("You should have received a copy of the GNU General Public License along with pppBenchmark.  If not, see http://www.gnu.org/licenses/.");

		} else {
			Map<String, Map<Double, int[]>> binGene = new HashMap<String, Map<Double, int[]>>();
			Map<String, Map<Double, int[]>> binTags = new HashMap<String, Map<Double, int[]>>();
			Map<String, Map<Double, int[]>> predGene = new HashMap<String, Map<Double, int[]>>();
			Map<String, Map<Double, int[]>> predTags = new HashMap<String, Map<Double, int[]>>();
			for (String file : parser.getRemainingArgs()) {
				System.out.println("Running 1B");
				binGene.put(file, aggregate(new BinGeneProtocol().analyze(new File(file))));
				System.out.println("Running 1A");
				binTags.put(file, aggregate(new BinTagsProtocol().analyze(new File(file))));
				System.out.println("Running 2B");
				predGene.put(file, aggregate(new PredGeneProtocol().analyze(new File(file))));
				System.out.println("Running 2A");
				predTags.put(file, aggregate(new PredTagsProtocol().analyze(new File(file))));
			}
			System.out.println("Merging 1B");
			XYSeriesCollection bg = graphBin(binGene);
			output(bg, "protocol 1B", prefix);
			System.out.println("Merging 1A");
			XYSeriesCollection bt = graphBin(binTags);
			output(bt, "protocol 1A", prefix);
			System.out.println("Merging 2B");
			XYSeriesCollection pg = graphPred(predGene);
			output(pg, "protocol 2B", prefix);
			System.out.println("Merging 2A");
			XYSeriesCollection pt = graphPred(predTags);
			output(pt, "protocol 2A", prefix);
		}

	}

	private static void parse(AutoHelpCmdLineParser parser, String[] args) {
		try {
			parser.parse(args);
		} catch (IllegalOptionValueException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (UnknownOptionException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}

	private static XYSeriesCollection graphBin(Map<String, Map<Double, int[]>> binTags) {
		XYSeriesCollection col = new XYSeriesCollection();

		for (String file : binTags.keySet()) {
			Map<Double, int[]> map = binTags.get(file);
			XYSeries xy = new XYSeries(file);
			double maxY = 0;
			double maxX = 0;
			double x = 0, y = 0;
			double lastX = 0;
			for (Double d : map.keySet()) {
				double tp = map.get(d)[0];
				double tn = map.get(d)[1];
				double fp = map.get(d)[2];
				double fn = map.get(d)[3];
				// if ((tp > 100 && fp > 100) || raw) {

				// if (tn != 0) {
				x = getPrecision(tp, tn, fp, fn);
				y = getRecall(tp, tn, fp, fn);
				// if (x >= lastX) {
				if (!Double.isNaN(x) && !Double.isNaN(y)) {
					xy.add(y, x);
					// if (y < 0.0001) {
					// System.out.println(line);
					// System.out.println(Config.predName(file) + "\t" +
					// x + "\t" + y);
					// }
					if (x > maxX)
						maxX = x;
					if (y > maxY)
						maxY = y;
					lastX = x;
					// } else {
					// System.err.println("Skip: " + d + "\t+" +
					// Arrays.toString(map.get(d))+"\t"+x+"\t"+y+"\t"+lastX);
					//							
					// }
					lastX = x;
				}
				// }else{
				// System.err.println("Not enough data.");
				// }

				// }

			}
			xy.add(0, maxX);
			if (maxY < 1) {
				xy.add(maxY + 0.00001, 0);
			}
			col.addSeries(xy);
			// System.out.println(file + "\t" + auprc(xy));
		}
		return col;
		// out.close();

	}

	private static void output(XYSeriesCollection col, String title, String prefix) throws FileNotFoundException {
		JFreeChart chart = ChartFactory.createXYLineChart("PRC-curve (" + title + ")", "recall", "precision", col, PlotOrientation.VERTICAL, true, false, false);
		XYPlot plot = (XYPlot) chart.getPlot();
		XYLineAndShapeRenderer xydot = new XYLineAndShapeRenderer();
		// xydot.setSeriesLinesVisible(12, false);
		// xydot.setSeriesLinesVisible(16, false);
		plot.setRenderer(xydot);
		// LegendTitle lg = chart.getLegend();
		// System.out.println(lg);
		PrintWriter out = new PrintWriter(prefix + ".auprc." + title + ".txt");
		out.println("#file\tauPRC");
		for (int i = 0; i < col.getSeriesCount(); i++) {
			XYSeries xy = col.getSeries(i);
			write(xy, prefix + "." + title + "." + xy.getKey());
			out.println(xy.getKey() + "\t" + auprc(xy));
		}
		out.close();
		GraphicsFileExport.exportPNG(new JFreeChartWrapper(chart), prefix + ".prccurve." + title, 1024, 800);

	}

	private static void write(XYSeries xy, String prefix) {

		try {
			PrintWriter out;
			out = new PrintWriter(prefix + ".detailed");

			for (int i = 0; i < xy.getItemCount(); i++) {
				out.println(xy.getX(i) + "\t" + xy.getY(i));

			}
			out.close();
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}

	private static XYSeriesCollection graphPred(Map<String, Map<Double, int[]>> predResults) {
		XYSeriesCollection col = new XYSeriesCollection();

		for (String file : predResults.keySet()) {
			Map<Double, int[]> map = predResults.get(file);
			XYSeries xy = new XYSeries(file);
			double firstX = -1000;
			double minY = 1000;
			double maxY = 0;
			double x = 0, y = 0;
			for (Double d : map.keySet()) {
				double b = map.get(d)[0];
				double a = map.get(d)[1];
				double c = map.get(d)[2];
				double e = map.get(d)[3];
				if (a != 0 && e != 0) {
					x = b / a;
					y = c / e;
					xy.add(y, x);
					if (y > maxY)
						maxY = y;
					if (y < minY) {
						minY = y;
						firstX = x;
					}

				}

			}
			xy.add(0, firstX);
			if (maxY < 1) {
				xy.add(maxY + 0.00001, 0);
			}
			col.addSeries(xy);
		}
		return col;
		// out.close();

	}

	private static Map<Double, int[]> aggregate(List<Map<Double, int[]>> list) {
		SortedMap<Double, int[]> total = new TreeMap<Double, int[]>();
		for (Map<Double, int[]> map : list) {
			for (Double key : map.keySet()) {
				if (!total.containsKey(key))
					total.put(key, new int[4]);
				for (int i = 0; i < 4; i++)
					total.get(key)[i] += map.get(key)[i];

			}
		}
		return total;
	}

	private static double getRecall(double tp, double tn, double fp, double fn) {
		return tp / (tp + fn);
	}

	private static double getPrecision(double tp, double tn, double fp, double fn) {
		return tp / (tp + fp);
	}

	private static double auprc(XYSeries xy) {
		double area = 0;
		for (int i = 0; i < xy.getItemCount() - 1; i++) {
			double xl = (Double) xy.getX(i + 1) - (Double) xy.getX(i);
			double yl = (Double) xy.getY(i + 1) / 2 + (Double) xy.getY(i) / 2;
			area += xl * yl;
		}
		return area;
	}
}
